##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

import numpy as np
import Tasmanian

def example_02():

    print("\n---------------------------------------------------------------------------------------------------\n")
    print("Example 2: integrate f(x,y) = exp(-x^2) * cos(y) over [-5,5] x [-2,3]")
    print("           using  Gauss-Patterson nodes and total degree polynomial space)")

    iNumDimensions = 2
    iExactness = 20

    fExactIntegral = 1.861816427518323e+00

    # the type_qptotal will guarantee exact integral for all polynomials with degree 20 or less
    grid = Tasmanian.makeGlobalGrid(iNumDimensions, 0, iExactness, "qptotal", "gauss-patterson")
    grid.setDomainTransform(np.array([[-5.0, 5.0], [-2.0, 3.0]])) # set the non-canonical domain

    aPoints = grid.getPoints()
    aWeights = grid.getQuadratureWeights()

    fApproximateIntegral = np.sum(aWeights * np.exp(-aPoints[:,0]**2) * np.cos(aPoints[:,1]))

    fError = np.abs(fApproximateIntegral - fExactIntegral)

    print("    at polynomial exactness".format(iExactness))
    print("    the grid has: {0:1d}".format(grid.getNumPoints()))
    print("    integral: {0:1.14e}".format(fApproximateIntegral))
    print("       error: {0:1.14e}\n".format(fError))

    iExactness = 40

    # the type_qptotal will guarantee exact integral for all polynomials with degree 20 or less
    grid = Tasmanian.makeGlobalGrid(iNumDimensions, 0, iExactness, "qptotal", "gauss-patterson")
    grid.setDomainTransform(np.array([[-5.0, 5.0], [-2.0, 3.0]])) # must reset the domain

    aPoints = grid.getPoints()
    aWeights = grid.getQuadratureWeights()

    fApproximateIntegral = np.sum(aWeights * np.exp(-aPoints[:,0]**2) * np.cos(aPoints[:,1]))

    fError = np.abs(fApproximateIntegral - fExactIntegral)

    print("    at polynomial exactness".format(iExactness))
    print("    the grid has: {0:1d}".format(grid.getNumPoints()))
    print("    integral: {0:1.14e}".format(fApproximateIntegral))
    print("       error: {0:1.14e}\n".format(fError))


if (__name__ == "__main__"):
    example_02()



##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

import numpy as np
import Tasmanian

def example_03():

    print("\n---------------------------------------------------------------------------------------------------\n")
    print("Example 3: integrate exp(-x1^2 - x2^2) * cos(x3) * cos(x4)")
    print("           for x1, x2 in [-5,5]; x3, x4 in [-2,3]")
    print("           using different rules and total degree polynomial space\n")

    def make_grid(iPrecision, sRule):
        grid = Tasmanian.makeGlobalGrid(4, 0, iPrecision, "qptotal", sRule)
        grid.setDomainTransform(np.array([[-5.0, 5.0], [-5.0, 5.0], [-2.0, 3.0], [-2.0, 3.0]]))
        return grid

    def print_error(grid):
        fExactIntegral = 1.861816427518323e+00 * 1.861816427518323e+00
        aPoints = grid.getPoints()
        aWeights = grid.getQuadratureWeights()

        fApproximateIntegral = np.sum(aWeights * np.exp(-aPoints[:,0]**2 -aPoints[:,1]**2)
                                      * np.cos(aPoints[:,2]) * np.cos(aPoints[:,3]))
        fError = np.abs(fApproximateIntegral - fExactIntegral)
        return "{0:>10d}{1:>10.2e}".format(grid.getNumPoints(), fError)

    print("               Clenshaw-Curtis      Gauss-Legendre    Gauss-Patterson")
    print(" precision    points     error    points     error    points    error")

    for prec in range(5, 41, 5):
        print("{0:>10d}{1:1s}{2:1s}{3:1s}".format(
            prec,
            print_error(make_grid(prec, "clenshaw-curtis")),
            print_error(make_grid(prec, "gauss-legendre-odd")),
            print_error(make_grid(prec, "gauss-patterson"))))

    print("\nAt 311K points the Gauss-Legendre error is O(1.E-1),")
    print("                   Clenshaw-Curtis error is O(1.E-7) at 320K points.")
    print("At 70K points the Gauss-Patterson error is O(1.E-4),")
    print("                  Clenshaw-Curtis needs 158K points to achieve the same.")


if (__name__ == "__main__"):
    example_03()



#!@Tasmanian_string_python_hashbang@

##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

from Tasmanian import DREAM
from Tasmanian import SparseGrid
from Tasmanian import makeGlobalGrid
from Tasmanian import makeSequenceGrid
from Tasmanian import loadNeededPoints
import numpy

def example_02():
    print("\n---------------------------------------------------------------------------------------------------\n")
    print("EXAMPLE 2: set the inference problem: identify model parameters x_0 and x_1")
    print("           from data (noise free example)")
    print("           model: f(x) = sin(x_0*M_PI*t + x_1),  data: d = sin(M_PI*t + 0.3*M_PI)")
    print("           t in [0,1], t is discretized with 32 equidistant nodes")
    print("           the likelihood is exp(- 16 * (f(x) - d)^2)")
    print("           using a sparse grid to interpolate the likelihood")
    print("     NOTE: See the comments in the C++ DREAM Example 2\n")

    iNumDimensions = 2;
    iNumChains = 100;
    iNumBurnupIterations = 3000;
    iNumCollectIterations = 100;

    def model(aX):
        # using 32 nodes in the interior of (0.0, 1.0)
        t = numpy.linspace(1.0 / 64.0, 1.0 - 1.0 / 64.0, 32)
        return numpy.sin(numpy.pi * aX[0] * t + aX[1])

    def likelihood(aY, aData):
        # using log-form likelihood
        # numpy.ones((1,)) converts the scalar to a 1D numpy array
        return - 0.5 * len(aData) * numpy.sum((aY - aData)**2) * numpy.ones((1,))

    aData = model([1.0, 0.3 * numpy.pi])

    domain_lower = numpy.array([0.5, -0.1])
    domain_upper = numpy.array([8.0,  1.7])

    grid = makeGlobalGrid(iNumDimensions, 1, 10, 'iptotal', 'clenshaw-curtis')
    grid.setDomainTransform(numpy.column_stack([domain_lower, domain_upper]))

    loadNeededPoints(lambda x, tid : likelihood(model(x), aData), grid, iNumThreads = 2)

    state = DREAM.State(iNumChains, grid)
    state.setState(DREAM.genUniformSamples(domain_lower, domain_upper, state.getNumChains()))

    DREAM.Sample(iNumBurnupIterations, iNumCollectIterations,
                 lambda x : grid.evaluateBatch(x).reshape((x.shape[0],)), # link to the SparseGrid
                 DREAM.Domain(grid),
                 state,
                 DREAM.IndependentUpdate("uniform", 0.5),
                 DREAM.DifferentialUpdate(90),
                 typeForm = DREAM.typeLogform)

    aMean, aVariance = state.getHistoryMeanVariance()

    print("Inferred values (using 10th order polynomial sparse grid):")
    print("    frequency:{0:13.6f}   error:{1:14.6e}".format(aMean[0],
                                                             numpy.abs(aMean[0] - 1.0)))
    print("   correction:{0:13.6f}   error:{1:14.6e}\n".format(aMean[1],
                                                             numpy.abs(aMean[0] - 0.3 * numpy.pi)))

    # repeat the example using the faster Sequence grid and 30th order polynomials
    grid = makeSequenceGrid(iNumDimensions, 1, 30, 'iptotal', 'leja')
    grid.setDomainTransform(numpy.column_stack([domain_lower, domain_upper]))

    loadNeededPoints(lambda x, tid : likelihood(model(x), aData), grid, iNumThreads = 2)

    state = DREAM.State(iNumChains, grid)
    state.setState(DREAM.genUniformSamples(domain_lower, domain_upper, state.getNumChains()))

    DREAM.Sample(iNumBurnupIterations, iNumCollectIterations,
                 lambda x : grid.evaluateBatch(x).reshape((x.shape[0],)), # link to the SparseGrid
                 DREAM.Domain(grid),
                 state,
                 DREAM.IndependentUpdate("uniform", 0.5),
                 DREAM.DifferentialUpdate(90),
                 typeForm = DREAM.typeLogform)

    aMean, aVariance = state.getHistoryMeanVariance()

    print("Inferred values (using 30th order polynomial sparse grid):")
    print("    frequency:{0:13.6f}   error:{1:14.6e}".format(aMean[0],
                                                             numpy.abs(aMean[0] - 1.0)))
    print("   correction:{0:13.6f}   error:{1:14.6e}".format(aMean[1],
                                                             numpy.abs(aMean[0] - 0.3 * numpy.pi)))

if __name__ == "__main__":
    example_02()



##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

from ctypes import c_char_p, c_int, c_double, c_void_p, POINTER, cdll, CFUNCTYPE
import numpy as np
import sys

from TasmanianConfig import __path_libdream__
from TasmanianConfig import enableVerboseErrors
from TasmanianConfig import TasmanianInputError as InputError

from TasmanianSG import TasmanianSparseGrid as SparseGrid

from TasmanianDreamLikely import *
from TasmanianDreamState import DreamState as State

type_dream_pdf     = CFUNCTYPE(None, c_int, c_int, POINTER(c_double), POINTER(c_double), POINTER(c_int))
type_dream_domain  = CFUNCTYPE(c_int, c_int, POINTER(c_double))
type_dream_iupdate = CFUNCTYPE(None, c_int, POINTER(c_double), POINTER(c_int))
type_dream_dupdate = CFUNCTYPE(c_double)
type_dream_random  = CFUNCTYPE(c_double)

pLibDTSG = cdll.LoadLibrary(__path_libdream__)

pLibDTSG.tsgGenUniformSamples.argtypes  = [c_int, c_int, POINTER(c_double), POINTER(c_double), c_char_p, c_int, type_dream_random]
pLibDTSG.tsgGenGaussianSamples.argtypes = [c_int, c_int, POINTER(c_double), POINTER(c_double), c_char_p, c_int, type_dream_random]

pLibDTSG.tsgDreamSample.argtypes = [c_int, c_int, c_int,
                                    type_dream_pdf, c_void_p,
                                    c_void_p, POINTER(c_double), POINTER(c_double), type_dream_domain,
                                    c_char_p, c_double, type_dream_iupdate,
                                    c_int, type_dream_dupdate,
                                    c_char_p, c_int, type_dream_random, POINTER(c_int)]


def pdfWrapper(callableProbability, iNumSamples, iNumDims, pX, pY, pErrInfo):
    '''
    Internal use only.

    Wraps around a user-provided callable method callableProbability
    to facilitate callback from the C/C++ code.
    The wrapper corresponds to a single output pdf/likelihood method.

    iNumSamples is the number of samples to process for this call.
    iNumDims is the number of dimensions of each sample vector
    pX is a POINTER(double) with size iNumSamples * iNumDims
    pY is a POINTER(double) with size iNumSamples
    '''
    pErrInfo[0] = 1
    aX = np.ctypeslib.as_array(pX, (iNumSamples, iNumDims))
    aY = np.ctypeslib.as_array(pY, (iNumSamples,))
    aResult = callableProbability(aX)
    if aResult.shape != (iNumSamples,) and enableVerboseErrors:
        print("ERROR: incorrect output from the probability distribution, should be (iNumSamples,)")
        return
    aY[0:iNumSamples] = aResult[0:iNumSamples]
    pErrInfo[0] = 0

def domainWrapper(callableDomain, iNumDims, pX):
    '''
    Internal use only.

    Wraps around a user-provided callable method callableDomain
    to facilitate callback from the C/C++ code.
    The wrapper corresponds to the test of a single point.

    iNumDims is the dimensions of the point
    pX is a POINTER(double) with size iNumDims
    returns 1 if callableDomain() returns true and 0 otherwise.
    '''
    aX = np.ctypeslib.as_array(pX, (iNumDims,))
    if callableDomain(aX):
        return 1;
    else:
        return 0

def iupdateWrapper(callableIUpdate, iNumDims, pX, pErrInfo):
    '''
    Internal use only.

    Wraps around a user-provided callable method callableIUpdate
    to facilitate callback from the C/C++ code.
    The wrapper corresponds to the independent update of a single point.

    iNumDims is the dimensions of the point
    pX is a POINTER(double) with size iNumDims
    pX will be overwritten with pX + update
    '''
    pErrInfo[0] = 1
    aX = np.ctypeslib.as_array(pX, (iNumDims,))
    aResult = callableIUpdate()
    if aResult.shape != (iNumDims,) and enableVerboseErrors:
        print("ERROR: incorrect output from the independent update, should be (iNumDimensions,)")
        return
    aX[0:iNumDims] = aResult[0:iNumDims]
    pErrInfo[0] = 0


class Domain(object):
    '''
    Domain objects specify the domain of sampling and is used to test
    each candidate sample. The objects roughly correspond to the inside()
    parameter given to TasDREAM::SampleDREAM() but with extra flexibility
    to reduce the cost of C++ to Python callbacks.

    The domain can be specified in one of the following ways:
    1) Domain(grid)
       The grid is an instance of Tasmanian.SparseGrid() and the domain
       will match the domain of the sparse grid. The grid domain is usually
       a hypercube with lower and upper limit, but in some cases, e.g.,
       using gauss-hermite and gauss-laguerre rules, the domain can be
       unbounded in some directions.
    2) Domain("hypercube", aLower, aUpper)
       The domain will be set to a hypercube, similar to TasDREAM::hypercube().
       The aLower and aUpper must be one dimensional numpy.ndarrays with
       size equal to the dimensions of the domain and the corresponding
       values must specify the lower and upper limit in each direction.
    3) Domain("unbounded")
       Samples will be accepted and rejected only based on the probability
       density, i.e., the sampling domain includes all real valued vectors.
    4) Domain("custom", inside)
       The inside parameter must be a callable object that accepts an single
       numpy.ndarray with size equal to the number of dimensions, and must
       return True or False depending on whether the vector is inside
       the domain.
    '''
    def __init__(self, sType, *args):
        '''
        Constructor that translates the input arguments into inputs
        to the internal Tasmanian API.

        See help(Tasmanian.DREAM.Domain)
        '''
        self.TasmanianDreamDomain = True
        if hasattr(sType, "TasmanianSparseGridObject"):
            self.pGrid = sType.pGrid
            self.pLower = None
            self.pUpper = None
            self.pCallable = type_dream_domain(lambda : 1)
        elif sType == "hypercube":
            self.pGrid = None
            self.pLower = np.ctypeslib.as_ctypes(args[0])
            self.pUpper = np.ctypeslib.as_ctypes(args[1])
            self.pCallable = type_dream_domain(lambda : 1)
        elif sType == "unbounded":
            self.pGrid = None
            self.pLower = None
            self.pUpper = None
            self.pCallable = type_dream_domain(lambda n, x : domainWrapper(lambda x: True, n, x))
        elif sType == "custom":
            self.pGrid = None
            self.pLower = None
            self.pUpper = None
            self.pCallable = type_dream_domain(lambda n, x : domainWrapper(args[0], n, x))
        else:
            raise InputError("Domain", "DREAM.Domain() was given an unknown domain type")


class IndependentUpdate(object):
    '''
    Each step of the DREAM procedure uses two updates to generate new proposed
    vectors for each chain. The independent update is a zero-mean random vector
    pulled from a distribution that does not depend on the other DREAM chains.
    Thus is the Python equivalent to the independent_update() parameter
    to TasDREAM::SampleDREAM().

    The independent update can be specified in one of the following ways:
    1) IndependentUpdate(sType, fMagnitude)
       The sType is a string and fMagnitude is a float64 corresponding to the
       dist and magnitude of the TasDREAM::SampleDREAM() overload.
       The sType must be either "gaussian" or "uniform".
       Note that the magnitude can be zero.
    2) IndependentUpdate(iupdate)
       The iupdate must be a callable object that accepts a numpy.ndarray
       with length equal to the sampling dimension and returns same vector
       but perturbed with the update, e.g., lambda x : x + update().
    '''
    def __init__(self, sType, callableOrMagnitude = 0.0):
        '''
        Constructor that translates the input arguments into inputs
        to the internal Tasmanian API.

        See help(Tasmanian.DREAM.IndependentUpdate)
        '''
        self.TasmanianDreamIndependentUpdate = True
        if not ((sys.version_info.major == 3 and isinstance(sType, str))
           or (sys.version_info.major == 2 and isinstance(sType, basestring))):
            if not callable(sType):
                raise InputError("sType", "DREAM.IndependentUpdate() the sType must be either a string or a callable object")
            self.sType = "null"
            self.fMagnitude = 0.0
            self.pCallable = type_dream_iupdate(lambda n, x, err : iupdateWrapper(sType, n, x, err))
        else:
            self.sType = sType
            self.fMagnitude = callableOrMagnitude
            self.pCallable = type_dream_iupdate(lambda : 1)
        if (sys.version_info.major == 3):
            self.sType = bytes(self.sType, encoding='utf8')


class DifferentialUpdate(object):
    '''
    Each step of the DREAM procedure uses two updates to generate new proposed
    vectors for each chain. The differential update is defined as a scale
    multiplied by the difference between two of the DREAM chains.
    This class sets the scale, i.e., the magnitude of the differential update.
    Thus is the Python equivalent to the differential_update() parameter
    to TasDREAM::SampleDREAM().

    The differential update can be specified in one of the following ways:
    1) DifferentialUpdate(iMagnitude)
       The iMagnitude is an integer that defines the percent scale of the
       differential update, i.e., the update is float(iMagnitude) / 100.0
       This is equivalent to using TasDREAM::const_percent<iMagnitude>
       in the C++ call to TasDREAM::SampleDREAM().
    2) DifferentialUpdate(dupdate)
       The dupdate must be a callable object that accepts no inputs
       and returns a float64 indicating the magnitude.

    Note: by symmetry, positive and negative scales with the same magnitude
          are statistically equivalent.
    '''
    def __init__(self, callableOrMagnitude):
        '''
        Constructor that translates the input arguments into inputs
        to the internal Tasmanian API.

        See help(Tasmanian.DREAM.DifferentialUpdate)
        '''
        self.TasmanianDreamDifferentialUpdate = True
        if (((sys.version_info.major == 3) and isinstance(callableOrMagnitude, int))
            or ((sys.version_info.major == 2) and isinstance(callableOrMagnitude, (int, long)))):
            self.iPercent = callableOrMagnitude
            self.pCallable = type_dream_dupdate(lambda : 1)
        else:
            if not callable(callableOrMagnitude):
                raise InputError("callableOrMagnitude", "DREAM.DifferentialUpdate() the callableOrMagnitude must be either an integer or a callable object")
            self.iPercent = -1
            self.pCallable = type_dream_dupdate(callableOrMagnitude)


class RandomGenerator(object):
    '''
    A class that helps specify the random number generator to be used on the
    C++ side of the call to TasDREAM::SampleDREAM().

    The random number engine can be defined in one of several ways:
    1) RandomGenerator("default") or just RandomGenerator()
       The C++ code will use the rand() command with seed initialized from
       the system time.
    2) RandomGenerator("minstd_rand")
       The rand() command in C++ is implementation dependent, i.e., the random
       number may be different between compilers even if the same seed is used.
       The std::minstd_rand generator is defined in the standard and the output
       is implementation independent, which makes it a good option for
       testing and debugging (probably not very good for actual applications).
       The random seed used here will still be taken from the system time.
    3) RandomGenerator("minstd_rand", iSeed)
       In addition to the standard std::minstd_rand engine this also specifies
       the seed.
    4) RandomGenerator(callableRNG = random01)
       The random01 parameter must be a callable object that returns random
       numbers distributed uniformly in (0, 1), e.g.,
       from random import uniform
       RandomGenerator(callableRNG = lambda : uniform(0.0, 1.0))
    '''
    def __init__(self, sType = "default", iSeed = -1, callableRNG = lambda : 1):
        '''
        Constructor that translates the input arguments into inputs
        to the internal Tasmanian API.

        See help(Tasmanian.DREAM.RandomGenerator)
        '''
        if sType not in ["default", "minstd_rand"]:
            raise InputError("sType", "DREAM.RandomGenerator() the sType is invalid")
        if not callable(callableRNG):
            raise InputError("callableRNG", "DREAM.RandomGenerator() the callableRNG must be a callable object")
        self.TasmanianDreamRandomGenerator = True
        self.sType = sType
        if (sys.version_info.major == 3):
            self.sType = bytes(self.sType, encoding='utf8')
        self.iSeed = iSeed
        self.pCallable = type_dream_random(callableRNG)

typePriorCallable = 0 # internal API, the values are synchronized with the C/C++ code
typePriorUniform  = 1

class Posterior(object):
    '''
    A class that helps combine the components of a Bayesian posterior into
    a single callable object. Strictly speaking, this is a factory class that
    generates probability distribution callable objects used in DREAM.Sample().
    The intended use of the class is embedded as in the C++ function
    TasDREAM::posterior().

    The probability distribution takes a two dimensional numpy.ndarray
    object with shape (num_samples, num_dimensions) and returns a one
    dimensional numpy.ndarray with length (num_samples,).
    Each entry of the result corresponds to the (unscaled) probability
    density at the corresponding vector.

    The usage of this class is:
    DREAM.Sample(..., Posterior(model, likelihood, prior), ...)
    or if log-form sampling is to be used:
    DREAM.Sample(..., Posterior(model, likelihood, prior, typeForm), ...)

    Combining the three components gives us:
        result = likelihood(typeForm, model(x)) * prior(x)
    of if using typeLogform:
        result = likelihood(typeForm, model(x)) + prior(x)

    The model, likelihood, and prior, can be callable objects so that

    model: input.shape  = (num_samples, num_dimensions)
           output.shape = (num_samples, num_outputs)

    likelihood: input.shape  = (num_samples, num_outputs)
                output.shape = (num_samples,)

    prior: input.shape  = (num_samples, num_dimensions)
           output.shape = (num_samples,)

    In addition:
    model: can be an instance of Tasmanian.SparseGrid()
    likelihood: can be an instance of DREAM.LikelihoodGaussIsotropic()
                of LikelihoodGaussAnisotropic().
    prior: can be the string "uniform"

    The typeForm must be either typeRegform (default) or typeLogform
    and must be synchronized with the form used in DREAM.Sample().
    The likelihood need only implement the form used in DREAM.Sample().
    The form here indicates whether to add or multiply by the log of
    the values of the prior.
    '''
    def __init__(self, model, likelihood, prior, typeForm = typeRegform):
        '''
        Constructor that translates the input arguments into inputs
        to the internal Tasmanian API.

        See help(Tasmanian.DREAM.Posterior)
        '''
        self.TasmanianPosterior = True

        self.typeForm = typeForm
        if self.typeForm not in [typeRegform, typeLogform]:
            raise InputError("typeForm", "unknown sampling form, must use typeRegform or typeLogform")

        self.model = model
        if hasattr(self.model, "TasmanianSparseGridObject"):
            self.bUseSparsegrid = True
        elif callable(self.model):
            self.bUseSparsegrid = False
        else:
            raise InputError("model", "model should be either an instance of Tasmanian.SparseGrid() or a callable object.")

        self.likelihood = likelihood
        if hasattr(self.likelihood, "TasmanianLikelihood"):
            self.bUseTasLikely = True
        elif callable(self.likelihood):
            self.bUseTasLikely = False
        else:
            raise InputError("likelihood", "likelihood should be either derived from TasmanianLikelihood() or a callable object.")

        if prior == "uniform":
            self.typePrior = typePriorUniform
        elif callable(prior):
            self.typePrior = typePriorCallable
            self.prior = prior
        else:
            raise InputError("prior", "prior should be either 'uniform' or a callable object.")

    def doModel(self, aX):
        '''
        Internal API, applies the model part of the posterior.
        '''
        if self.bUseSparsegrid:
            return self.model.evaluateBatch(aX)
        else:
            return self.model(aX)

    def doLikelihood(self, aModelValues):
        '''
        Internal API, applies the likelihood part of the posterior.
        '''
        if self.bUseTasLikely:
            return self.likelihood.getLikelihood(self.typeForm, aModelValues)
        else:
            return self.likelihood(self.typeForm, aModelValues)

    def doPrior(self, aLikely, aX):
        '''
        Internal API, computes the prior part of the posterior.
        '''
        if self.typePrior == typePriorUniform:
            aResult = aLikely
        else:
            # reg-form multiplies, log-form adds
            if self.typeForm == typeRegform:
                aResult = aLikely * self.prior(aX)
            else:
                aResult = aLikely + self.prior(aX)
        return aResult

    def distribution(self):
        '''
        Internal API, combines the input objects into a single lambda.
        '''
        return lambda x : self.doPrior(self.doLikelihood(self.doModel(x)), x).reshape((x.shape[0],))

def PosteriorEmbeddedLikelihood(model, prior, typeForm = typeRegform):
    '''
    Shortcut to Tasmanian.DREAM.Posterior() when the model and the likelihood
    are combined into a single object (callable or sparse grid).
    The prior and typeForm are the same.
    '''
    return Posterior(model, lambda t, x : x, prior, typeForm)

def genUniformSamples(lfLower, lfUpper, iNumSamples, random01 = RandomGenerator(sType = "default")):
    '''
    Wrapper around TasDREAM::genUniformSamples()

    Using the one dimensional numpy.ndarrays lfLower and lfUpper,
    which describe the upper and lower bounds of a hypercube,
    generate iNumSamples from a uniform distribution using
    the random01 engine.
    The lengths of lfLower and lfUpper must match.

    Returns a two dimensional numpy.ndarray of shape = (iNumSamples, len(lfLower)),
    '''
    if len(lfLower) != len(lfUpper):
        raise InputError("lfUpper", "The length of lfLower and lfUpper do not match.")

    aLower = np.array(lfLower)
    aUpper = np.array(lfUpper)
    iNumDims = len(lfLower)

    aResult = np.empty((iNumDims * iNumSamples,), np.float64)

    pLibDTSG.tsgGenUniformSamples(iNumDims, iNumSamples, np.ctypeslib.as_ctypes(aLower), np.ctypeslib.as_ctypes(aUpper),
                                  c_char_p(random01.sType), random01.iSeed, random01.pCallable,
                                  np.ctypeslib.as_ctypes(aResult))
    return aResult.reshape((iNumSamples, iNumDims))


def genGaussianSamples(lfMean, lfDeviation, iNumSamples, random01 = RandomGenerator(sType = "default")):
    '''
    Wrapper around TasDREAM::tsgGenGaussianSamples()

    Using the one dimensional numpy.ndarrays lfMean and lfVariance,
    which describe the mean and standard deviation of a Gaussian distribution,
    generate iNumSamples from a normal (Gaussian) distribution using
    the random01 engine.
    The lengths of lfMean and lfDeviation must match.

    Returns a two dimensional numpy.ndarray of shape = (iNumSamples, len(lfLower)),
    '''
    if len(lfMean) != len(lfDeviation):
        raise InputError("lfDeviation", "The length of lfMean and lfDeviation do not match.")

    aMean = np.array(lfMean)
    aDeviation = np.array(lfDeviation)
    iNumDims = len(lfMean)

    aResult = np.empty((iNumDims * iNumSamples,), np.float64)

    pLibDTSG.tsgGenGaussianSamples(iNumDims, iNumSamples, np.ctypeslib.as_ctypes(aMean), np.ctypeslib.as_ctypes(aDeviation),
                                   c_char_p(random01.sType), random01.iSeed, random01.pCallable,
                                   np.ctypeslib.as_ctypes(aResult))
    return aResult.reshape((iNumSamples, iNumDims))


def Sample(iNumBurnup, iNumCollect,
           probability_distibution, domain_description, dream_state,
           independent_update = IndependentUpdate("none"),
           differential_update = DifferentialUpdate(100),
           random01 = RandomGenerator(sType = "default"),
           typeForm = typeRegform):
    '''
    Wrapper to TasDREAM::SampleDREAM().
    The Markov Chain Monte Carlo algorithms generate chains of samples
    with distribution that converges to the target probability density.
    The current state of the chains (i.e., the last vector) is updated by
    first generating a set of proposals and then accepting/rejecting randomly
    but with the constraint that the only accepted vectors belong to the
    specified domain. The proposals are constructed from two components,
    the independent component that is iid random with zero mean and
    the differential component that is takes from the scaled difference
    between the state of two randomly selected chains.

    iNumBurnup: is a non-negative integer that indicates the number of
                iterations to discard, i.e., samples that have no statistical
                significance because the chain state has not converged to
                the target probability distribution.
    iNumCollect: is a non-negative integer that indicates the number of
                iterations to store in the state history.

    probability_distibution: is either an instance of DREAM.Posterior() or
                a callable object that accepts a two dimensional numpy.ndarray
                and then returns a one dimensional numpy.ndarray.
                See help(Tasmanian.DREAM.Posterior)
                Can also use a lambda to capture a Tasmanian sparse grid, i.e.,
                lambda x : grid.evaluateBatch(x).reshape((x.shape[0],))

    domain_description: is an instance of DREAM.Domain()
                See help(Tasmanian.DREAM.Domain)

    dream_state: is an instance of DREAM.State() with number of dimensions
                that match the dimensions used by the probability_distibution.
                See help(Tasmanian.DREAM.State)
                On input the state chains have to be initialized, on output
                the history will be populated with the iNumCollect samples.
                (if the history is not empty, the new samples will be appended)

    independent_update: is an instance of DREAM.IndependentUpdate()
                See help(Tasmanian.DREAM.IndependentUpdate)

    differential_update: is an instance of DREAM.DifferentialUpdate()
                See help(Tasmanian.DREAM.DifferentialUpdate)

    random01: is an instance of DREAM.RandomGenerator()
                See help(Tasmanian.DREAM.RandomGenerator)

    typeForm: is either DREAM.typeRegform or DREAM.typeLogform
               The form modifies the accept/reject criteria based on whether
               the probability_distibution returns the actual distribution
               of interest or the logarithm of that (which is sometimes more
               stable and/or easier to approximate with a sparse grid).

    Note: the typeForm must match the return of the probability_distibution.
    '''
    if not hasattr(domain_description, "TasmanianDreamDomain"):
        raise InputError("domain_description", "domain_description must be an instance of DREAM.Domain()")
    if not hasattr(dream_state, "TasmanainDreamState"):
        raise InputError("dream_state", "dream_state must be an instance of DREAM.State()")
    if not hasattr(independent_update, "TasmanianDreamIndependentUpdate"):
        raise InputError("independent_update", "independent_update must be an instance of DREAM.IndependentUpdate()")
    if not hasattr(differential_update, "TasmanianDreamDifferentialUpdate"):
        raise InputError("differential_update", "differential_update must be an instance of DREAM.DifferentialUpdate()")
    if not hasattr(random01, "TasmanianDreamRandomGenerator"):
        raise InputError("random01", "random01 must be an instance of DREAM.RandomGenerator()")
    if typeForm not in [typeRegform, typeLogform]:
        raise InputError("typeForm", "unknown sampling form, must use typeRegform or typeLogform")

    pErrorCode = (c_int * 1)()

    if hasattr(probability_distibution, "TasmanianPosterior"):
        pLibDTSG.tsgDreamSample(typeForm, c_int(iNumBurnup), c_int(iNumCollect),
                                type_dream_pdf(lambda m, n, x, y, err : pdfWrapper(probability_distibution.distribution(), m, n, x, y, err)),
                                dream_state.pStatePntr,
                                domain_description.pGrid, domain_description.pLower, domain_description.pUpper, domain_description.pCallable,
                                c_char_p(independent_update.sType), independent_update.fMagnitude, independent_update.pCallable,
                                differential_update.iPercent, differential_update.pCallable,
                                c_char_p(random01.sType),
                                random01.iSeed,
                                random01.pCallable,
                                pErrorCode)
    elif callable(probability_distibution):
        pLibDTSG.tsgDreamSample(typeForm, c_int(iNumBurnup), c_int(iNumCollect),
                                type_dream_pdf(lambda m, n, x, y, err : pdfWrapper(probability_distibution, m, n, x, y, err)), dream_state.pStatePntr,
                                domain_description.pGrid, domain_description.pLower, domain_description.pUpper, domain_description.pCallable,
                                c_char_p(independent_update.sType), independent_update.fMagnitude, independent_update.pCallable,
                                differential_update.iPercent, differential_update.pCallable,
                                c_char_p(random01.sType),
                                random01.iSeed,
                                random01.pCallable,
                                pErrorCode)
    else:
        raise InputError("probability_distibution", "probability_distibution must be a callable object that takes a 2D numpy.ndarray and returns a 1D ndarray")

    if pErrorCode[0] != 0:
        raise InputError("Sample", "An error occurred during the call to Tasmanian.")




import unittest
import TasmanianSG
import os
import numpy as np

import testCommon

ttc = testCommon.TestTasCommon()

class TestTasClass(unittest.TestCase):
    '''
    Miscelaneous tests that don't quite fit in other categories.
    '''
    def __init__(self):
        unittest.TestCase.__init__(self, "testNothing")

    def testNothing(self):
        pass

    def checkPolynomialSpace(self):
        '''
        Construct grids with specific parameters and check the polynomial
        space against known (pen-and-paper) exact spaces.
        '''
        grid = TasmanianSG.TasmanianSparseGrid()

        aA = np.array([[0, 0], [0, 1], [0, 2], [0, 3], [0, 4], [1, 0], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2], [3, 0], [4, 0]])
        grid.makeGlobalGrid(2, 1, 2, 'level', 'clenshaw-curtis')
        aP = grid.getGlobalPolynomialSpace(True)
        np.testing.assert_equal(aP, aA, "poly space mismatch", True)

        aA = np.array([[0, 0], [0, 1], [0, 2], [0, 3], [0, 4], [0, 5], [1, 0], [1, 1], [1, 2], [1, 3], [1, 4], [1, 5], [2, 0], [2, 1], [2, 2], [2, 3], [3, 0], [3, 1], [3, 2], [3, 3], [4, 0], [4, 1], [5, 0], [5, 1]])
        grid.makeGlobalGrid(2, 1, 2, 'level', 'clenshaw-curtis')
        aP = grid.getGlobalPolynomialSpace(False)
        np.testing.assert_equal(aP, aA, "poly space mismatch", True)

        aA = np.array([[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [2, 0]])
        grid.makeSequenceGrid(2, 1, 2, 'level', 'rleja')
        aP = grid.getGlobalPolynomialSpace(True)
        np.testing.assert_equal(aP, aA, "poly space mismatch", True)

        grid.makeSequenceGrid(2, 1, 2, 'level', 'rleja')
        aP = grid.getGlobalPolynomialSpace(False)
        np.testing.assert_equal(aP, aA, "poly space mismatch", True)

        aA = np.array([[0, 0], [0, 1], [0, 2], [0, 3], [1, 0], [1, 1], [1, 2], [1, 3], [2, 0], [2, 1], [3, 0], [3, 1]])
        grid.makeGlobalGrid(2, 1, 1, 'level', 'leja-odd')
        aP = grid.getGlobalPolynomialSpace(False)
        np.testing.assert_equal(aP, aA, "poly space mismatch", True)

        grid.makeSequenceGrid(2, 1, 2, 'level', 'leja')
        aP = grid.getGlobalPolynomialSpace(False)
        np.testing.assert_equal(aP, aA, "poly space mismatch", True)

    def checkPlotting(self):
        '''
        If matplotlib is available and there is an active display, then
        make plot commands but do not show the final plots.
        If matplotlib is missing, check the exceptions.
        '''
        grid = TasmanianSG.TasmanianSparseGrid()

        if (TasmanianSG.bTsgPlotting):
            if ((os.name == "posix") and (os.environ.get('DISPLAY') is None)):
                print("NOTE: there is no display, cannot run plotting tests.")
                return
            grid.makeGlobalGrid(2, 1, 20, 'level', 'leja')
            ttc.loadExpN2(grid)
            grid.plotResponse2D()
            grid.plotPoints2D()

            grid.makeGlobalGrid(2, 1, 0, 'level', 'leja')
            ttc.loadExpN2(grid)
            grid.plotResponse2D()
            grid.plotPoints2D()

            grid.makeGlobalGrid(3, 1, 3, 'level', 'leja')
            try:
                grid.plotPoints2D()
                self.assertTrue(False, "failed to flag plot exception when when using a grid with other than 2D")
            except TasmanianSG.TasmanianInputError as TSGError:
                TSGError.bShowOnExit = False
                self.assertEqual(TSGError.sVariable, "plotPoints2D", "error raising exception for plotPoints2D() using test\n Error.sVariable = '{0:1s}'".format(TSGError.sVariable))
            try:
                grid.plotResponse2D()
                self.assertTrue(False, "failed to flag plot exception when when using a grid with other than 2D")
            except TasmanianSG.TasmanianInputError as TSGError:
                TSGError.bShowOnExit = False
                self.assertEqual(TSGError.sVariable, "plotResponse2D", "error raising exception for plotResponse2D() using test\n Error.sVariable = '{0:1s}'".format(TSGError.sVariable))

            grid.makeGlobalGrid(2, 2, 2, 'level', 'leja')
            try:
                grid.plotResponse2D(iOutput = -1)
                self.assertTrue(False, "failed to flag plot exception when when using wrong output")
            except TasmanianSG.TasmanianInputError as TSGError:
                TSGError.bShowOnExit = False
                self.assertEqual(TSGError.sVariable, "iOutput", "error raising exception for plotResponse2D() using test\n Error.sVariable = '{0:1s}'".format(TSGError.sVariable))
            try:
                grid.plotResponse2D(iOutput = 3)
                self.assertTrue(False, "failed to flag plot exception when when using wrong output")
            except TasmanianSG.TasmanianInputError as TSGError:
                TSGError.bShowOnExit = False
                self.assertEqual(TSGError.sVariable, "iOutput", "error raising exception for plotResponse2D() using test\n Error.sVariable = '{0:1s}'".format(TSGError.sVariable))
            try:
                grid.plotResponse2D(iNumDim0 = -1)
                self.assertTrue(False, "failed to flag plot exception when when using wrong output")
            except TasmanianSG.TasmanianInputError as TSGError:
                TSGError.bShowOnExit = False
                self.assertEqual(TSGError.sVariable, "iNumDim0", "error raising exception for plotResponse2D() using test\n Error.sVariable = '{0:1s}'".format(TSGError.sVariable))
            try:
                grid.plotResponse2D(iNumDim1 = -1)
                self.assertTrue(False, "failed to flag plot exception when when using wrong output")
            except TasmanianSG.TasmanianInputError as TSGError:
                TSGError.bShowOnExit = False
                self.assertEqual(TSGError.sVariable, "iNumDim1", "error raising exception for plotResponse2D() using test\n Error.sVariable = '{0:1s}'".format(TSGError.sVariable))
        else:
            grid.makeGlobalGrid(2, 1, 3, 'level', 'leja')
            ttc.loadExpN2(grid)
            try:
                grid.plotResponse2D()
                self.assertTrue(False, "failed to flag plotexception  when missing matplotlib")
            except TasmanianSG.TasmanianInputError as TSGError:
                TSGError.bShowOnExit = False
                self.assertEqual(TSGError.sVariable, "plotResponse2D", "error raising exception for plotResponse2D() using test\n Error.sVariable = '{0:1s}'".format(TSGError.sVariable))
            try:
                grid.plotPoints2D()
                self.assertTrue(False, "failed to flag plot exception when missing matplotlib")
            except TasmanianSG.TasmanianInputError as TSGError:
                TSGError.bShowOnExit = False
                self.assertEqual(TSGError.sVariable, "plotPoints2D", "error raising exception for plotPoints2D() using test\n Error.sVariable = '{0:1s}'".format(TSGError.sVariable))

    def performMiscTests(self):
        self.checkPolynomialSpace()
        self.checkPlotting()



import unittest
import TasmanianSG
import numpy as np

from random import shuffle

import testCommon

ttc = testCommon.TestTasCommon()

class TestTasClass(unittest.TestCase):
    '''
    Test the refinement capabilities:
    * set different refinements
    * estimate anisotropic coefficients
    * read/write refinements
    '''
    def __init__(self):
        unittest.TestCase.__init__(self, "testNothing")

    def testNothing(self):
        pass

    def checkSetClear(self):
        '''
        Set refinement and clear refinement
        '''
        grid = TasmanianSG.TasmanianSparseGrid()
        grid.makeLocalPolynomialGrid(2, 1, 2, 1, 'semi-localp')
        ttc.loadExpN2(grid)
        self.assertEqual(grid.getNumNeeded(), 0, "num needed")
        grid.setSurplusRefinement(0.0001, 0, 'classic')
        self.assertTrue((grid.getNumNeeded() > 0), "num needed")
        grid.clearRefinement()
        self.assertEqual(grid.getNumNeeded(), 0, "num needed")

    def checkAnisoCoeff(self):
        '''
        Check anisotropic coefficients
        '''
        grid = TasmanianSG.TasmanianSparseGrid()
        grid.makeGlobalGrid(2, 1, 9, 'level', 'rleja')
        aP = grid.getPoints()
        aV = np.exp(aP[:,0] + aP[:,1]**2)
        grid.loadNeededPoints(aV.reshape([aP.shape[0], 1]))
        aC = grid.estimateAnisotropicCoefficients('iptotal', 0)
        self.assertEqual(len(aC.shape), 1, 'dimensions of the estimated anisotropic weight')
        self.assertEqual(aC.shape[0], 2, 'dimensions of the estimated anisotropic weight')
        self.assertLess(np.abs(float(aC[0]) / float(aC[1]) - 2.0), 0.2, 'wrong anisotropic weights estimated')
        aC = grid.estimateAnisotropicCoefficients('ipcurved', 0)
        self.assertEqual(len(aC.shape), 1, 'dimensions of the estimated anisotropic weight')
        self.assertEqual(aC.shape[0], 4, 'dimensions of the estimated anisotropic weight')
        self.assertLess(np.abs(float(aC[0]) / float(aC[1]) - 2.0), 0.2, 'wrong anisotropic weights estimated, alpha curved')
        self.assertLess(aC[2], 0.0, 'wrong anisotropic weights estimated, beta 1')
        self.assertLess(aC[3], 0.0, 'wrong anisotropic weights estimated, beta 2')

    def checkLocalpSurplus(self):
        '''
        Check surplus refinement for local polynomial grids
        '''
        grid = TasmanianSG.TasmanianSparseGrid()
        grid.makeLocalPolynomialGrid(2, 1, 4, 1, 'semi-localp')
        ttc.loadExpN2(grid)
        aPoints = grid.getPoints()
        aScale = np.array([[1.0 if aPoints[i,0] > 0.0 else 0.0] for i in range(aPoints.shape[0])])
        grid.setSurplusRefinement(1.E-9, 0, 'classic', [], aScale)
        aNeeded = grid.getNeededPoints()
        for iI in range(aNeeded.shape[0]):
            self.assertLess(0.0, aNeeded[iI, 0], 'wrong set of needed points after rescaling')


    def checkFileIO(self):
        '''
        Read/Write regular refinement.
        '''
        grid = TasmanianSG.TasmanianSparseGrid()

        sRefinementIOGrids = ["grid.makeGlobalGrid(2, 1, 2, 'level', 'clenshaw-curtis')",
                              "grid.makeSequenceGrid(2, 1, 2, 'level', 'rleja')"]
        for sTest in sRefinementIOGrids:
            exec(sTest)
            ttc.loadExpN2(grid)
            grid.setAnisotropicRefinement('iptotal', 20, 0)
            self.assertTrue(grid.getNumNeeded() > 0, 'did not refine')
            gridB = TasmanianSG.TasmanianSparseGrid()
            gridB.copyGrid(grid)
            ttc.compareGrids(grid, gridB)

            grid.write("refTestFlename.grid", bUseBinaryFormat = True)
            gridB.makeLocalPolynomialGrid(1, 1, 0, 1)
            gridB.read("refTestFlename.grid")
            ttc.compareGrids(grid, gridB)

            grid.write("refTestFlename.grid", bUseBinaryFormat = False)
            gridB.makeLocalPolynomialGrid(1, 1, 0, 1)
            gridB.read("refTestFlename.grid")
            ttc.compareGrids(grid, gridB)

    def checkConstruction(self):
        '''
        Test read/write when using construction.
        '''
        llTest = ["gridA.makeGlobalGrid(3, 2, 2, 'level', 'clenshaw-curtis'); gridB.makeGlobalGrid(3, 2, 2, 'level', 'clenshaw-curtis')",
                  "gridA.makeSequenceGrid(3, 2, 4, 'level', 'leja'); gridB.makeSequenceGrid(3, 2, 4, 'level', 'leja')",
                  "gridA.makeLocalPolynomialGrid(3, 2, 2); gridB.makeLocalPolynomialGrid(3, 2, 2)",
                  "gridA.makeWaveletGrid(3, 2, 2); gridB.makeWaveletGrid(3, 2, 2)",
                  "gridA.makeFourierGrid(3, 2, 2, 'level'); gridB.makeFourierGrid(3, 2, 2, 'level')",]

        for sMakeGrids in llTest:
            for sFormat in [False, True]: # test binary and ascii format
                gridA = TasmanianSG.TasmanianSparseGrid()
                gridB = TasmanianSG.TasmanianSparseGrid()
                gridC = TasmanianSG.TasmanianSparseGrid()

                exec(sMakeGrids)

                gridA.beginConstruction()
                gridB.beginConstruction()
                #gridA.printStats()

                gridB.write("testSave", bUseBinaryFormat = sFormat)
                gridB.makeSequenceGrid(1, 1, 0, "level", "rleja") # clean the grid
                gridB.read("testSave")
                ttc.compareGrids(gridA, gridB)
                gridC.copyGrid(gridA)
                ttc.compareGrids(gridA, gridC)

                for t in range(5): # use 5 iterations
                    if (gridA.isLocalPolynomial() or gridA.isWavelet()):
                        aPointsA = gridA.getCandidateConstructionPointsSurplus(1.E-4, "fds")
                        aPointsB = gridB.getCandidateConstructionPointsSurplus(1.E-4, "fds")
                        aPointsC = gridC.getCandidateConstructionPointsSurplus(1.E-4, "fds")
                    else:
                        aPointsA = gridA.getCandidateConstructionPoints("level", 0)
                        aPointsB = gridB.getCandidateConstructionPoints("level", 0)
                        aPointsC = gridC.getCandidateConstructionPoints("level", 0)
                    np.testing.assert_almost_equal(aPointsA, aPointsB, decimal=11)
                    np.testing.assert_almost_equal(aPointsA, aPointsC, decimal=11)

                    iNumPoints = int(aPointsA.shape[0] / 2)
                    if (iNumPoints > 32): iNumPoints = 32

                    # use the first samples (up to 32) and shuffle the order
                    # add one of the samples further in the list
                    liSamples = list(range(iNumPoints + 1))
                    shuffle(liSamples)
                    for iI in range(len(liSamples)):
                        if (liSamples[iI] == iNumPoints):
                            liSamples[iI] = iNumPoints + 1
                    #liSamples = map(lambda i: i if i < iNumPoints else iNumPoints + 1, liSamples)

                    for iI in liSamples: # compute and load the samples
                        aPoint = aPointsA[iI, :]
                        aValue = np.array([np.exp(aPoint[0] + aPoint[1]), 1.0 / ((aPoint[0] - 1.3) * (aPoint[1] - 1.6) * (aPoint[2] - 2.0))])

                        gridA.loadConstructedPoint(aPoint, aValue)
                        gridB.loadConstructedPoint(aPoint, aValue)
                        gridC.loadConstructedPoint(aPoint, aValue)

                    # using straight construction or read/write should produce the same result
                    ttc.compareGrids(gridA, gridC)
                    gridB.write("testSave", bUseBinaryFormat = sFormat)
                    gridB.makeSequenceGrid(1, 1, 0, "level", "rleja")
                    gridB.read("testSave")
                    ttc.compareGrids(gridA, gridB)
                    gridC.copyGrid(gridA)
                    ttc.compareGrids(gridA, gridC)

                gridA.finishConstruction()
                gridB.finishConstruction()

                gridB.write("testSave", bUseBinaryFormat = sFormat)
                gridB.makeSequenceGrid(1, 1, 0, "level", "rleja")
                gridB.read("testSave")
                ttc.compareGrids(gridA, gridB)
                gridC.copyGrid(gridA)
                ttc.compareGrids(gridA, gridC)

        # check multi-point load
        gridA = TasmanianSG.TasmanianSparseGrid()
        gridA.makeLocalPolynomialGrid(3, 2, 4);
        ttc.loadExpN2(gridA)

        gridB = TasmanianSG.TasmanianSparseGrid()
        gridB.makeLocalPolynomialGrid(3, 2, 0)

        gridB.beginConstruction()
        aX = gridA.getPoints()
        aY = gridA.evaluateBatch(aX)
        gridB.loadConstructedPoint(aX, aY)
        gridB.finishConstruction()
        ttc.compareGrids(gridA, gridB)

        # check some mem-leaks and crashes (correctness is elsewhere)
        gridA = TasmanianSG.TasmanianSparseGrid()
        gridA.makeLocalPolynomialGrid(2, 5, 0)
        gridA.beginConstruction()
        gridA.loadConstructedPoint(np.empty([0, 2]), np.empty([0, 5])) # empty input, check for crash

        gridA.makeLocalPolynomialGrid(2, 1, 1)
        gridA.loadNeededPoints(np.ones([5, 1]))
        gridA.beginConstruction()
        aPoints = gridA.getCandidateConstructionPointsSurplus(1.E-4, "classic") # should generate empty output
        np.testing.assert_almost_equal(aPoints, np.empty([0, 0]), 14, "failed to generate empty list of construction points", True)

        gridA.makeLocalPolynomialGrid(2, 1, 0)
        gridA.loadNeededPoints(np.ones([1, 1]))
        gridA.beginConstruction()
        aPoints = gridA.getCandidateConstructionPointsSurplus(1.E-4, "classic", 0, [], np.array([[1.E-6]])) # should generate empty output
        np.testing.assert_almost_equal(aPoints, np.empty([0, 0]), 14, "failed to generate empty list of construction points", True)

        gridA.makeGlobalGrid(2, 1, 1, "tensor", "clenshaw-curtis")
        gridA.loadNeededPoints(np.ones([9, 1]))
        gridA.beginConstruction()
        aPoints = gridA.getCandidateConstructionPoints("ipcurved", [5, 5, 2, 2], [1, 1]) # should generate empty output
        np.testing.assert_almost_equal(aPoints, np.empty([0, 0]), 14, "failed to generate empty list of construction points", True)

    def checkRemovePoints(self):
        '''
        tests removePointsByHierarchicalCoefficient()
        '''
        grid = TasmanianSG.makeLocalPolynomialGrid(2, 1, 1)
        aPoints = grid.getNeededPoints()
        grid.loadNeededValues(np.exp(-aPoints[:,0]**2 -0.5*aPoints[:,1]**2).reshape((grid.getNumNeeded(), 1)))

        reduced = TasmanianSG.TasmanianSparseGrid()

        reduced.copyGrid(grid)
        reduced.removePointsByHierarchicalCoefficient(0.6)
        self.assertEqual(reduced.getNumPoints(), 3, "failed to remove points with threshold 0.6")
        np.testing.assert_almost_equal(reduced.getLoadedPoints(), np.array([[0.0, 0.0], [-1.0, 0.0], [1.0, 0.0]]), 14, "failed reduce 1", True)

        reduced.copyGrid(grid)
        reduced.removePointsByHierarchicalCoefficient(0.7)
        self.assertEqual(reduced.getNumPoints(), 1, "failed to remove points with threshold 0.7")
        np.testing.assert_almost_equal(reduced.getLoadedPoints(), np.array([[0.0, 0.0]]), 14, "failed reduce 2", True)

        reduced.copyGrid(grid)
        reduced.removePointsByHierarchicalCoefficient(0.0, iNumKeep = 3)
        self.assertEqual(reduced.getNumPoints(), 3, "failed to remove points down to 3")
        np.testing.assert_almost_equal(reduced.getLoadedPoints(), np.array([[0.0, 0.0], [-1.0, 0.0], [1.0, 0.0]]), 14, "failed reduce 3", True)

        reduced.copyGrid(grid)
        reduced.removePointsByHierarchicalCoefficient(0.0, iNumKeep = 1)
        self.assertEqual(reduced.getNumPoints(), 1, "failed to remove points down to 1")
        np.testing.assert_almost_equal(reduced.getLoadedPoints(), np.array([[0.0, 0.0]]), 14, "failed reduce 4", True)

        reduced.copyGrid(grid)
        reduced.removePointsByHierarchicalCoefficient(0.0, aScaleCorrection = np.array([[1.0], [1.0], [1.0], [0.1], [0.1]]), iNumKeep = 3)
        self.assertEqual(reduced.getNumPoints(), 3, "failed to remove corrected points")
        np.testing.assert_almost_equal(reduced.getLoadedPoints(), np.array([[0.0, 0.0], [0.0, -1.0], [0.0, 1.0]]), 14, "failed reduce 5", True)

    def performRefinementTest(self):
        self.checkSetClear()
        self.checkAnisoCoeff()
        self.checkLocalpSurplus()
        self.checkFileIO()
        self.checkConstruction()
        self.checkRemovePoints()



import unittest
import Tasmanian
import os
import numpy as np

import testCommon

ttc = testCommon.TestTasCommon()

def batch_exp(x):
    aResult = np.zeros((x.shape[0], 1))
    for i in range(x.shape[0]):
        aResult[i, 0] = np.exp(-np.sum(x[i,:]))
    return aResult

class TestTasClass(unittest.TestCase):
    '''
    Miscelaneous tests that don't quite fit in other categories.
    '''
    def __init__(self):
        unittest.TestCase.__init__(self, "testNothing")

    def testNothing(self):
        pass

    def checkLoadNeeded(self):
        '''
        Check the load needed construction procedure.
        '''
        gridA = Tasmanian.SparseGrid()
        gridB = Tasmanian.SparseGrid()

        for i in range(4):
            if i < 2:
                gridA.makeGlobalGrid(2, 1, 3, 'level', 'fejer2')
                gridB.copyGrid(gridA)
                Tasmanian.loadNeededValues(lambda x, i : np.ones((1,)) * np.exp(-np.sum(x**2)), gridA, i)
            else:
                gridA.makeLocalPolynomialGrid(2, 1, 2, 2)
                gridB.copyGrid(gridA)
                gridA.printStats()
                gridA.loadNeededPoints(np.ones((gridA.getNumPoints(), 1)))
                Tasmanian.reloadLoadedPoints(lambda x, i : np.ones((1,)) * np.exp(-np.sum(x**2)), gridA, i - 2)

            ttc.loadExpN2(gridB)
            ttc.compareGrids(gridA, gridB)

    def checkConstruction(self):
        gridA = Tasmanian.SparseGrid()
        gridB = Tasmanian.SparseGrid()

        for i in range(12):
            if i < 2:
                gridA.makeGlobalGrid(2, 1, 3, 'level', 'fejer2')
                gridB.copyGrid(gridA)
                Tasmanian.constructAnisotropicSurrogate(lambda x, i : np.ones((1,1)) * np.exp(-np.sum(x**2)),
                                                        gridA.getNumPoints(), i, 1, gridA, "hyperbolic", 0)
            elif i < 4:
                gridA.makeGlobalGrid(2, 1, 3, 'level', 'min-lebesgue')
                gridB.copyGrid(gridA)
                Tasmanian.constructAnisotropicSurrogate(lambda x, y, i : np.ones((1,1)) * np.exp(-np.sum(x**2)),
                                                        gridA.getNumPoints(), i - 2, 1, gridA, "ipcurved", (6, 5, 1, 1),
                                                        bUseInitialGuess = True, sCheckpointFilename = "cpf_addons")
                os.remove("cpf_addons")
            elif i < 6:
                gridA.makeGlobalGrid(2, 1, 3, 'level', 'min-delta', liLevelLimits = (5, 5))
                gridB.copyGrid(gridA)
                gridA.clearLevelLimits()
                Tasmanian.constructAnisotropicSurrogate(lambda x, i : np.ones((1,1)) * np.exp(-np.sum(x**2)),
                                                        gridA.getNumPoints(), i - 4, 1, gridA, "iplevel", (1, 2),
                                                        liLevelLimits = (5, 5), bUseInitialGuess = False)
            elif i < 8:
                gridA.makeGlobalGrid(2, 1, 3, 'level', 'min-delta', liLevelLimits = (10, 10))
                gridB.copyGrid(gridA)
                gridA.clearLevelLimits()
                Tasmanian.constructAnisotropicSurrogate(lambda x, y, i : np.ones((1,1)) * np.exp(-np.sum(x**2)),
                                                        gridA.getNumPoints(), i - 6, 1, gridA, "iplevel", 0,
                                                        liLevelLimits = (10, 10), bUseInitialGuess = True)
            elif i < 10:
                gridA.makeLocalPolynomialGrid(2, 1, 2, 2, liLevelLimits = (7, 7))
                gridB.copyGrid(gridA)
                gridA.clearLevelLimits()
                Tasmanian.constructSurplusSurrogate(lambda x, y, i : np.ones((1,1)) * np.exp(-np.sum(x**2)),
                                                    gridA.getNumPoints(), i - 8, 1, gridA, 1.E-5, "stable", -1,
                                                    liLevelLimits = (7, 7), bUseInitialGuess = True)
            elif i < 12:
                gridA.makeLocalPolynomialGrid(2, 1, 2, 3)
                gridB.copyGrid(gridA)
                Tasmanian.constructSurplusSurrogate(lambda x, i : np.ones((1,1)) * np.exp(-np.sum(x**2)),
                                                    gridA.getNumPoints(), i - 10, 1, gridA, 1.E-5, "fds", 0,
                                                    bUseInitialGuess = False, sCheckpointFilename = "cpf_addons")
                os.remove("cpf_addons")

            ttc.loadExpN2(gridB)
            gridA.finishConstruction()
            ttc.compareGrids(gridA, gridB)

    def checkBatchConstruct(self):
        gridA = Tasmanian.SparseGrid()
        gridB = Tasmanian.SparseGrid()

        gridA.makeGlobalGrid(2, 1, 6, 'level', 'fejer2')
        gridB.copyGrid(gridA)
        Tasmanian.constructAnisotropicSurrogate(lambda x, i : batch_exp(x),
                                                gridA.getNumPoints(), 2, 1, gridA, "iptotal", 0)
        Tasmanian.constructAnisotropicSurrogate(lambda x, i : batch_exp(x),
                                                gridB.getNumPoints(), 2, 10, gridB, "iptotal", 0)
        ttc.compareGrids(gridA, gridB)

        gridA.makeLocalPolynomialGrid(2, 1, 3)
        gridB.copyGrid(gridA)
        Tasmanian.constructSurplusSurrogate(lambda x, i : batch_exp(x),
                                            gridA.getNumPoints(), 2, 1, gridA, 1.E-11, "classic")
        Tasmanian.constructSurplusSurrogate(lambda x, i : batch_exp(x),
                                            gridB.getNumPoints(), 2, 10, gridB, 1.E-11, "classic")
        ttc.compareGrids(gridA, gridB)

    def checkUnstructuredL2(self):
        grid = Tasmanian.makeLocalPolynomialGrid(2, 1, 5)
        if not grid.isAccelerationAvailable("cpu-blas"): return

        x = np.linspace(-0.9, 0.9, 40)
        xx, yy = np.meshgrid(x, x)

        points = np.column_stack([xx.reshape((xx.size,)), yy.reshape((yy.size,))])
        model = np.exp(points[:,0] - points[:,1]).reshape((points.shape[0], 1))

        Tasmanian.loadUnstructuredDataL2(points, model, 1.E-5, grid)

        x = np.linspace(-0.8, 0.8, 20)
        xx, yy = np.meshgrid(x, x)
        points = np.column_stack([xx.reshape((xx.size,)), yy.reshape((yy.size,))])
        model = np.exp(points[:,0] - points[:,1])

        surrogate = grid.evaluateBatch(points)

        self.assertLess(np.max(np.abs(model - surrogate[:,0])), 5.E-3, "Constructed from unstructured too inaccurate.")

    def checkExoticQuadrature(self):
        # Initialize some useful constants and utility functions.
        integrand = lambda x : np.exp(-np.linalg.norm(x) ** 2)
        sinc1s0 = lambda x : np.sinc(x / np.pi)
        sinc10s1 = lambda x : np.sinc(10 * (x - 1) / np.pi)
        max_level = 25
        nref = 101
        def create_surrogate(lambda_fn):
            grid = Tasmanian.makeGlobalGrid(1, 1, nref, "level", "gauss-legendre")
            grid.loadNeededValues(np.apply_along_axis(lambda_fn, 1, grid.getNeededPoints()))
            return grid
        def compute_integral(grid, integrand):
            return np.sum(np.apply_along_axis(integrand, 1, grid.getPoints()) * grid.getQuadratureWeights())
        # Each pair below consists of a 1D integral value (first) and a CustomTabulated object (second) used to approximate this integral.
        testPairs = [
            [1.4321357541271255E-00, Tasmanian.createExoticQuadratureFromFunction(max_level, 0.0, sinc1s0, nref, "Sinc1s0-shift0-Symm-byFn", True)],
            [1.4321357541271255E-00, Tasmanian.createExoticQuadratureFromFunction(max_level, 0.0, sinc1s0, nref, "Sinc1s0-shift0-nonSymm-byFn", False)],
            [1.4321357541271255E-00, Tasmanian.createExoticQuadratureFromFunction(max_level, 1.0, sinc1s0, nref, "Sinc1s0-shift1-Symm-byFn", True)],
            [6.4062055930705356E-02, Tasmanian.createExoticQuadratureFromFunction(max_level, 1.0, sinc10s1, nref, "Sinc10s1-shift1-nonSymm-byFn", False)],
            [1.4321357541271255E-00, Tasmanian.createExoticQuadratureFromGrid(max_level, 0.0, create_surrogate(sinc1s0), "Sinc1s0-shift0-Symm-byGrid", True)],
            [1.4321357541271255E-00, Tasmanian.createExoticQuadratureFromGrid(max_level, 0.0, create_surrogate(sinc1s0), "Sinc1s0-shift0-nonSymm-byGrid", False)],
            [1.4321357541271255E-00, Tasmanian.createExoticQuadratureFromGrid(max_level, 1.0, create_surrogate(sinc1s0), "Sinc1s0-shift1-Symm-byGrid", True)],
            [6.4062055930705356E-02, Tasmanian.createExoticQuadratureFromGrid(max_level, 1.0, create_surrogate(sinc10s1), "Sinc10s1-shift1-nonSymm-byGrid", False)],
        ]
        # Test the accuracy of exotic quadrature.
        for integral, ct in testPairs:
            self.assertEqual(max_level, ct.getNumLevels(), "Loaded number of levels does not match actual number of levels!")
            grid = Tasmanian.makeGlobalGridCustom(1, 0, max_level-1, "level", ct)
            np.testing.assert_almost_equal(compute_integral(grid, integrand), integral, 10,
                                           "Computed integral does not match exact integral in test instance " + ct.getDescription())
            grid.makeGlobalGridCustom(2, 0, max_level-1, "level", ct)
            np.testing.assert_almost_equal(compute_integral(grid, integrand), integral ** 2, 10,
                                           "Computed integral does not match exact integral in test instance " + ct.getDescription())

    def performAddonTests(self):
        self.checkLoadNeeded()
        self.checkConstruction()
        self.checkBatchConstruct()
        self.checkUnstructuredL2()
        self.checkExoticQuadrature()
